#include "graphics.h"

/* returns position of <x>,<y>'th element in a single dimensional array
 * with <num_col> columns 
*/
int pos(int x, int y, int num_col){
	return y * num_col + x;
}

/* returns position of <x>,<y>'th element in a single dimensional array
 * with <num_col> columns. Where the center element is indexed as 0,0
 * and negative indexs go back and up.
 */
int c_pos(int x, int y, int rad){
	x = x + rad;
	y = y + rad;
	return pos(x, y, 2*rad+1);
}

/* returns pixel <x>,<y> from <image> or NULL if location is outside of the image
*/
unsigned char* getPixel(TargaImage* image, int x, int y){
	if(x < 0 || x >= image->width || y < 0 || y >= image->height) {
		return NULL;
	} else {
		return image->data + pos(x, y, image->width) * 4;
	}
}

/*copies <src> pixel to <dest> pixel
*/
void copy_pxl(unsigned char* src, unsigned char* dest){
	for(int i = 0; i < 4; i++){
		dest[i] = src[i];
	}
}

/* resizes <image> to <new_width>x<new_heigth> pixels using the given <filter>
 * for resampling
 */
TargaImage* resize(TargaImage* image, int new_width, int new_height, float(*filter)(float x, float r)){	
	//create new image for horizontal resize
	unsigned char* horz_data = new unsigned char[new_width * image->height * 4];
	TargaImage* horz_image = new TargaImage(new_width, image->height, horz_data);
	//and another for the vertical resize
	unsigned char* vert_data = new unsigned char[new_width * new_height * 4];
	TargaImage* vert_image = new TargaImage(new_width, new_height, vert_data);
	
	resample_horz(image, horz_image, filter);
	resample_vert(horz_image, vert_image, filter);

	//remove the useless
	delete horz_image;

	return vert_image;
}

/* samples <new_image> from a reconstructed <image> vertically using the given <filter>
 * for resampling
 */
void resample_vert(TargaImage* image, TargaImage* new_image, float(*filter)(float, float)){	
	float delta_y	= (float)image->height/new_image->height;
	float rad;
	//choose the correct radius
	if(delta_y > 1){
		rad		= delta_y;
	} else {
		rad		= l;
	}

	float y0		= delta_y/2.0 - 0.5;

	//loop over the image
	for(int i = 0; i < image->width; i++){
		for(int j = 0; j < new_image->height; j++){
			unsigned char* new_pxl = getPixel(new_image, i, j);
			new_pxl[0] = new_pxl[1] = new_pxl[2] = new_pxl[3] = 0;
			
			double norm = 0;
			float y = y0 + j * delta_y;

			for(int k = ceil(y - rad); k < floor(y + rad); k++){
				unsigned char* pxl = getPixel(image, i, k);

				if(pxl != NULL){
					float f_val = filter(y - k, rad);
					norm += f_val;
					//sum averaged weight
					new_pxl[0] += pxl[0] * f_val;
					new_pxl[1] += pxl[1] * f_val;
					new_pxl[2] += pxl[2] * f_val;
				}
			}
			//renormalize
			if(norm != 0){
				new_pxl[0] /= norm;
				new_pxl[1] /= norm;
				new_pxl[2] /= norm;
			}
		}
	}
}

/*
*/
void convolute2d_pxl(TargaImage* image, unsigned char* new_pxl, float x, float y, float(*filter)(float, float), int x_rad, int y_rad){
	new_pxl[0] = new_pxl[1] = new_pxl[2] = new_pxl[3] = 0;
	double norm = 0;

	for(int i = ceil(x - x_rad); i <= floor(x + x_rad); i++){
		for(int j = ceil(y - y_rad); j <= floor(y + y_rad); j++){
			unsigned char* pxl = getPixel(image, i, j);

			if(pxl != NULL){
				float f_val = filter(x - i, x_rad) * filter(y - j, y_rad);
				norm += f_val;
				
				//sum averaged weight
				new_pxl[0] += pxl[0] * f_val;
				new_pxl[1] += pxl[1] * f_val;
				new_pxl[2] += pxl[2] * f_val;
			}
		}
	}
	//renormalize
	if(norm != 0){
		new_pxl[0] /= norm;
		new_pxl[1] /= norm;
		new_pxl[2] /= norm;
	}
}

/* samples <new_image> from a reconstructed <image> horizontally using the given <filter>
 * for resampling
 */
void resample_horz(TargaImage* image, TargaImage* new_image, float(*filter)(float, float)){	
	float delta_x	= (float)image->width/new_image->width;
	float rad;
	//choose the correct radius
	if(delta_x > 1){
		rad		= delta_x;
	} else {
		rad		= l;
	}

	float x0		= delta_x/2.0 - 0.5;
	
	//loop over the image
	for(int j = 0; j < image->height; j++){
		for(int i = 0; i < new_image->width; i++){
			unsigned char* new_pxl = getPixel(new_image, i, j);
			new_pxl[0] = new_pxl[1] = new_pxl[2] = new_pxl[3] = 0;
			//printf("new_pxl(%d,%d) ", i, j);
			
			double norm = 0;
			float x = x0 + i * delta_x;

			for(int k = ceil(x - rad); k < floor(x + rad); k++){
				//printf("y:%lf (%d,%d) ", y, i, k);
				unsigned char* pxl = getPixel(image, k, j);

				if(pxl != NULL){
					float f_val = filter(x - k, rad);
					norm += f_val;
					//sum averaged weight
					new_pxl[0] += pxl[0] * f_val;
					new_pxl[1] += pxl[1] * f_val;
					new_pxl[2] += pxl[2] * f_val;
				}
			}
			//renormalize
			if(norm != 0){
				new_pxl[0] /= norm;
				new_pxl[1] /= norm;
				new_pxl[2] /= norm;
			}
		}
	}
}

/* Dithers pixel <x>,<y> of <image> and diffuses the error according to <d_matrix>,
 *  a square matrix of size <d_size>
*/
void FS_dither_pxl(TargaImage* image, int x, int y, float* d_matrix, int d_size){
	float error;
	unsigned char* err_pxl;
	unsigned char* pxl = getPixel(image, x, y);

	//thresh-hold the current pixel
	if(pxl[0] < 128){
		error = (float)pxl[0];
		pxl[0] = 0;
		pxl[1] = 0;
		pxl[2] = 0;

	} else {
		error = (float)pxl[0] - 255;
		pxl[0] = 255;
		pxl[1] = 255;
		pxl[2] = 255;
	}

	//distribute the error
	for(int i = 0; i < d_size; i++){
		for(int j = 0; j < d_size; j++){
			if((err_pxl = getPixel(image, x + (i - d_size/2), y + (j - d_size/2))) != NULL){
				float dis_err = d_matrix[pos(i, j, d_size)] * error;
			
				if(err_pxl[0] + dis_err > 255) {
					err_pxl[0] = 255;
				} else if(err_pxl[0] + dis_err < 0) {
					err_pxl[0] = 0;
				} else {
					err_pxl[0] += (unsigned char)dis_err;
				}
			}
		}
	}
}

//returns a black and white version of the <image>
TargaImage* bw(TargaImage* image){
	float d_right[] = {	0,		0,		0,
						0,		0,	.4375,
					.1875,	.3125,	.0625};

	float d_left[] = {	0,		0,		0,
					.4375,		0,		0,
					.0625,	.3125,	.1875};
	
	//get new grey scaled version of image
	TargaImage* bw_image = greyscale(image);

	//run Floyd-Steinburg Algorithm
	for(int y=0; y < image->height; y++) {		
		if(y%2 == 0){
			for(int x=0; x < image->width; x++){
				FS_dither_pxl(bw_image, x, y, d_right, 3);
			}
		} else {
			for(int x=image->width-1; x >= 0; x--){
				FS_dither_pxl(bw_image, x, y, d_left, 3);
			}
		}
	}

	return bw_image;
}

//returns a greyscaled version of the <image>
TargaImage* greyscale(TargaImage* image){
	//create new image
	unsigned char* new_data = new unsigned char[image->width * image->height * 4];
	TargaImage* grey_image = new TargaImage(image->width, image->height, new_data);
	
	//greyscale the image
	for(int x=0; x < image->width; x++){
		for(int y=0; y < image->height; y++){
			unsigned char* pxl  = getPixel(image, x,y);
			unsigned char* pxl_grey = getPixel(grey_image, x, y);
				int L = .3 * pxl[0] + .59 * pxl[1] + .11 * pxl[2];	//Y = 0.30R + 0.59G + 0.11B
				pxl_grey[0]	= L;		// desaturate red
				pxl_grey[1]	= L;		// desaturate green
				pxl_grey[2]	= L;		// desaturate blue
				pxl_grey[3]	= pxl[3];	// leave alpha alone
		}
	}
	return grey_image;
}

/* places <brush> stroke of radius <brush_rad> on the <new_image> given the value of the corresponding
 * pixel <x>,<y> in the original <image>.
 */
void impress_pxl(TargaImage* image, TargaImage* new_image, int x, int y, float* brush, int brush_rad){
	unsigned char* pxl  = getPixel(image, x, y);
	for(int i = -brush_rad; i < brush_rad; i++){
		for(int j = -brush_rad; j < brush_rad; j++){
			unsigned char* new_pxl  = getPixel(new_image, x+i, y+j);
			if(new_pxl != NULL){
				new_pxl[0] = brush[c_pos(i, j, brush_rad)] * pxl[0] + (1 - brush[c_pos(i, j, brush_rad)]) * new_pxl[0];
				new_pxl[1] = brush[c_pos(i, j, brush_rad)] * pxl[1] + (1 - brush[c_pos(i, j, brush_rad)]) * new_pxl[1];
				new_pxl[2] = brush[c_pos(i, j, brush_rad)] * pxl[2] + (1 - brush[c_pos(i, j, brush_rad)]) * new_pxl[2];
			}
		}
	}
}

/* returns a painted verison of <image> using the selected brush
 */
TargaImage* impress(TargaImage* image, int brush){
	const int SC_BRUSH = 0;
	const int C_BRUSH  = 1;
	const int SP_BRUSH = 2;
	const int BS_BRUSH = 3;
	const int TL_BRUSH = 4;
	const int ALL_BRUSH = 5;
	int all_prob = 30;

	//small circle brush
	int sc_prob = 10;
	int sc_brush_rad = 3;
	float sc_brush[] =
					{	0,	0,	1,	1,	1,	0,	0,
						0,	1,	1,	1,	1,	1,	0,
						1,	1,	1,	1,	1,	1,	1,
						1,	1,	1,	1,	1,	1,	1,
						1,	1,	1,	1,	1,	1,	1,
						0,	1,	1,	1,	1,	1,	0,
						0,	0,	1,	1,	1,	0,	0};
	
	//cresent brush
	int c_prob = 6;
	int c_brush_rad = 3;
	float c_brush[] =
					{	0,	1,	1,	.3,	.2,	.1,	0,
						0,	1,	1,	0,	0,	0,	0,
						1,	1,	0,	0,	0,	0,	0,
						1,	1,	0,	0,	0,	0,	0,
						1,	1,	0,	0,	0,	0,	0,
						0,	1,	0,	0,	0,	0,	0,
						0,	1,	1,	1,	1,	.4,	0};
	//spike brush
	//cresent brush
	int sp_prob = 4;
	int sp_brush_rad = 3;
	float sp_brush[] =
					{	0,	0,	0,	0,	0,	1,	0,
						0,	0,	0,	0,	1,	0,	0,
						0,	0,	1,	1,	1,	0,	0,
						0,	1,	1,	1,	0,	0,	0,
						1,	1,	1,	1,	0,	0,	0,
						1,	1,	1,	0,	0,	0,	0,
						1,	1,	1,	0,	0,	0,	0};

	//big square brush
	int bs_prob = 500;
	int bs_brush_rad = 20;
	float *bs_brush = new float[1681];
	for(int i = 0; i < 1681; i++){
		bs_brush[i] = 1;
	}

	//three lines brush
	int tl_prob = 14;
	int tl_brush_rad = 7;
	float tl_brush[] =
					{	.1,	.2,	.3,	.4,	.4,	.5,	.6,	.6, .6,	.5,	.4,	.4,	.3,	.2,	.1,
						.2,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	.2,	
						.1,	.2,	.3,	.4,	.4,	.5,	.6,	.6,	.6,	.5,	.4,	.4,	.3,	.2,	.1,
						0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,
						0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,
						0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,
						.1,	.2,	.3,	.4,	.4,	.5,	.6,	.6,	.6,	.5,	.4,	.4,	.3,	.2,	.1,
						.2,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	.2,
						.1,	.2,	.3,	.4,	.4,	.5,	.6,	.6,	.6,	.5,	.4,	.4,	.3,	.2,	.1,
						0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	
						0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,
						0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,
						.1,	.2,	.3,	.4,	.4,	.5,	.6,	.6,	.6,	.5,	.4,	.4,	.3,	.2,	.1,
						.2,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	.2,	
						.1,	.2,	.3,	.4,	.4,	.5,	.6,	.6,	.6,	.5,	.4,	.4,	.3,	.2,	.1};

	//create new image
	unsigned char* new_data = new unsigned char[image->width * image->height * 4];
	TargaImage* new_image = new TargaImage(image->width, image->height, new_data);

	
	if(brush != ALL_BRUSH){
		for(int y=0; y < new_image->height; y++){
			for(int x=0; x < new_image->width; x++){
				switch(brush){
					case SC_BRUSH:
						if(rand()%sc_prob == 0)	impress_pxl(image, new_image, x, y, sc_brush, sc_brush_rad);
						break;
					case C_BRUSH:
						if(rand()%c_prob == 0)	impress_pxl(image, new_image, x, y, c_brush, c_brush_rad);
						break;
					case SP_BRUSH:
						if(rand()%sp_prob == 0)	impress_pxl(image, new_image, x, y, sp_brush, sp_brush_rad);
						break;
					case BS_BRUSH:
						if(rand()%bs_prob == 0)	impress_pxl(image, new_image, x, y, bs_brush, bs_brush_rad);
						break;
					case TL_BRUSH:
						if(rand()%tl_prob == 0)	impress_pxl(image, new_image, x, y, tl_brush, tl_brush_rad);
						break;
					default:
						cout << "IMPRESSIONIST usage: p1 <input_image> <output_image> 2 <brush_number(0-5)>" << endl;
						exit(1);
				}
			}
		}
	} else {
		//run all brushes in series
		//big square
		for(int y=0; y < new_image->height; y++){
			for(int x=0; x < new_image->width; x++){
				if(rand()%(bs_prob*1000)) impress_pxl(image, new_image, x, y, bs_brush, bs_brush_rad);
			}
		}
		//small circle
		for(int y=0; y < new_image->height; y++){
			for(int x=0; x < new_image->width; x++){
				if(rand()%(sc_prob*20) == 0) impress_pxl(image, new_image, x, y, sc_brush, sc_brush_rad);
			}
		}
		//cresent
		for(int y=0; y < new_image->height; y++){
			for(int x=0; x < new_image->width; x++){
				if(rand()%(c_prob*20) == 0) impress_pxl(image, new_image, x, y, c_brush, c_brush_rad);
			}
		}
		//spike
		for(int y=0; y < new_image->height; y++){
			for(int x=0; x < new_image->width; x++){
				if(rand()%(sp_prob*20) == 0) impress_pxl(image, new_image, x, y, sp_brush, sp_brush_rad);
			}
		}
		//threee lines
		for(int y=0; y < new_image->height; y++){
			for(int x=0; x < new_image->width; x++){
				if(rand()%(tl_prob*20) == 0) impress_pxl(image, new_image, x, y, tl_brush, tl_brush_rad);
			}
		}
	}

	return new_image;
}

/* returns a number normalized to the range [-1, 1] given
 * a number in [0,max]
 */
double normalize(int x, int max){
	float MD2 = max/2.0;
	return x/MD2 - 1;
}

/* returns a number normalized to the range [0, max] given
 * a number in [-1,1]
 */
float denormalize(double x, int max){
	float MD2 = max/2.0;
	return (1 + x) * MD2;
}

//nl-transform on <x>,<y>
void warp0_pxl(TargaImage* image, float* x, float* y){
	int x_max = image->width - 1;
	int y_max = image->height - 1;
	double x_norm = normalize(*x, x_max);
	double y_norm = normalize(*y, y_max);
	double r = sqrt(x_norm*x_norm + y_norm*y_norm);
	double phi = atan2(y_norm, x_norm);
	
	*x = denormalize(asin(x_norm)/PID2, x_max);
	*y = denormalize(asin(y_norm)/PID2, y_max);
}

//nl-transform on <x>,<y>
void warp1_pxl(TargaImage* image, float* x, float* y){
	int x_max = image->width - 1;
	int y_max = image->height - 1;
	double x_norm = normalize(*x, x_max);
	double y_norm = normalize(*y, y_max);
	double r = sqrt(x_norm*x_norm + y_norm*y_norm);
	double phi = atan2(y_norm, x_norm);
	r = sqrt(r);

	*x = denormalize(r * cos(phi), x_max);
	*y = denormalize(r * sin(phi), y_max);
}

//nl-transform on <x>,<y>
void warp2_pxl(TargaImage* image, float* x, float* y){
	int x_max = image->width - 1;
	int y_max = image->height - 1;
	double x_norm = normalize(*x, x_max);
	double y_norm = normalize(*y, y_max);
	double r = sqrt(x_norm*x_norm + y_norm*y_norm);
	double phi = atan2(y_norm, x_norm);
	r = 2*asin(r)/PI;

	*x = denormalize(r * cos(phi), x_max);
	*y = denormalize(r * sin(phi), y_max);
}


/* returns a warped image using the non-linear transform described by <transform>
 * using <filter> for interpolation
*/
TargaImage* warp(TargaImage* image, void(*transform)(TargaImage*, float*, float*), int filter){
	//create new image
	unsigned char* new_data = new unsigned char[image->width * image->height * 4];
	TargaImage* new_image = new TargaImage(image->width, image->height, new_data);

	unsigned char* pxl;
	unsigned char* new_pxl;
	for(int y=0; y < new_image->height; y++){
		for(int x=0; x < new_image->width; x++){
			float u = x, v = y;
			transform(image, &u, &v);

			if(filter == 0){//no interpolation
				pxl = getPixel(image, (int)u, (int)v);
				new_pxl = getPixel(new_image, x, y);
				if(pxl != NULL){
					copy_pxl(pxl, new_pxl);
				} else {
					new_pxl[0] = 0;
					new_pxl[1] = 0;
					new_pxl[2] = 0;
				}

			} else if(filter == 1){//fixed size interpolation radius of 3
				new_pxl = getPixel(new_image, x, y);
				convolute2d_pxl(image, new_pxl, u, v, &bspline_filter, 3, 3);
	
			} else if(filter == 2) {//variable size interpolation
				new_pxl = getPixel(new_image, x, y);

				float px_u = x - 1, px_v = y;
				float nx_u = x + 1, nx_v = y;
				float py_u = x, py_v = y - 1;
				float ny_u = x, ny_v = y + 1;

				transform(image, &px_u, &px_v);
				transform(image, &nx_u, &nx_v);
				transform(image, &py_u, &py_v);
				transform(image, &ny_u, &ny_v);

				float x_rad = abs(nx_u - px_u);
				float y_rad = abs(ny_v - py_v);

				if(x_rad < l) x_rad = l;
				if(y_rad < l) y_rad = l;
	
				//if(x_rad > 0 && y_rad > 0) cout << x_rad << " " << y_rad << " " ;
				convolute2d_pxl(image, new_pxl, u, v, &bspline_filter, x_rad, y_rad);
			} else {
				cout << "WARP usage: p1 <input_image> <output_image> 3 <warp_number(0-2)> <filter_number(0-2)>" << endl;
				exit(1);
			}
		}
	}

	return new_image;
}


